package combinatorParsing

import CharArrayReader._
import collection.mutable.HashSet

/** This class augments generic parsers by a tokenizing functionality.
 *  It should be used as a base class for all parsers operating on arrays of characters
 *  as inputs.
 */
trait TokenizingParsers extends GenericParsers {

  /** The delimiters used for the tokenizer
   *  A delimiter is a symbol which always constitutes a single character
   *  token, even if not preceded or followed by whitespace.
   *  Inheriting classes are supposed to fill this set, using <code>incl</code>;
   *  Example:
   *  <pre>
   *     delimiters.incl('(', ')', ';')
   *  </pre>
   */
  protected val delimiters = new HashSet[char]

  /** The keywords used for the tokenizer
   *  Strings matching a keyword yield KW tokens instead of Identifier tokens.
   *  Inheriting classes are supposed to fill this set, using <code>incl</code>;
   *  Example:
   *  <pre>
   *     keywords.incl("if", "then", "else")
   *  </pre>
   *  <p>
   *    Delimiter tokens are always keywords, no need to add their names to
   *    the <code>keywords</code> set.
   *  </p>
   */
  protected val keywords = new HashSet[String]

  /** Utility function to characterize whitespace characters */
  protected def isWhiteSpace(ch: char) = ch <= ' ' && ch != EofCh

  /** Utility function to characterize digits */
  protected def isDigit(ch: char) = '0' <= ch && ch <= '9'

  /** Utility function to characterize delimiters */
  protected def isDelimiter(ch: char) = delimiters contains ch

  /** The main method of this class. Creates an iterator yielding the tokens
   *  of a source array of characters.
   */
  def tokenize(source: Array[char]) = new Iterator[Token] {
    private val in = new CharArrayReader(source)
    in.next

    private var atEnd = false

    def hasNext: boolean = !atEnd

    def next: Token = {
      while (isWhiteSpace(in.ch)) in.next
      val chars = new StringBuffer
      val pos = Position(in.line, in.column, source)
      if (isDelimiter(in.ch)) {
        chars append in.ch; in.next
        KW(chars.toString)(pos)
      } else if (in.ch == '\"' || in.ch == '\'') {
        val delim = in.ch; in.next
        while (in.ch != EofCh && in.ch != NewLineCh && in.ch != delim) {
          chars append in.ch; in.next
        }
        if (in.ch == delim) { 
          in.next
          StringLit(chars.toString)(pos) 
        } else {
          ErrorToken("unclosed string literal")(pos) 
        }
      } else if (isDigit(in.ch)) {
        do {
          chars append in.ch; in.next
        } while (isDigit(in.ch))
        NumericLit(chars.toString)(pos)
      } else if (in.ch == EofCh) {
        atEnd = true
        EOF()(pos)
      } else {
        do {
          chars append in.ch; in.next
        } while (in.ch != EofCh && !isWhiteSpace(in.ch) && !isDelimiter(in.ch))
        val name = chars.toString
        if (keywords contains name) KW(name)(pos)
        else Identifier(name)(pos)
      }
    }
  }

  /** Run phrase(p) on given source */
  def run(p: Parser, source: Array[char]): ParseResult =
    phrase(p)(tokenize(source).toList)

  /** Run phrase(p) on given source */
  def run(p: Parser, source: String): ParseResult = run(p, source.toCharArray)
}


